#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "newsLib.h"
#include <map>
#include <algorithm>

using namespace std;

typedef map<int, pair<int, int> > MII;

MII mm;

#define MX 15
#define REALLY_HUGE 100000

int x,y;
char ar[MX][MX];

int encode(int sx, int sy, int tl, int tr, int bl, int br) {

	if (sx < 0) sx = 0;
	if (sx < 0) sy = 0;
	int var = 0;
	var |= (sx & 0xF);
	var <<= 4;
	var |= (sy & 0xF);
	var <<= 4;
	var |= (tl & 0xF);
	var <<= 4;
	var |= (tr & 0xF);
	var <<= 4;
	var |= (bl & 0xF);
	var <<= 4;
	var |= (br & 0xF);

	//printf("enc(%d,%d,%d,%d,%d,%d)=%x\n", sx,sy,tl,tr,bl,br,var);
	return var;
}

int canonar(int *mx, int *my) {
	int minx,miny,maxx,maxy;
	int var = 0;

	minx = miny = MX+1;
	maxx = maxy = -1;
	for (int xx = 0; xx < x; xx++) {
		for (int yy = 0; yy < y; yy++) {
			if (ar[yy][xx] == 'X') {
				minx=min(minx, xx);
				miny=min(miny, yy);
				maxx=max(maxx, xx);
				maxy=max(maxy, yy);
			}
		}
	}

	int tl,tr,bl,br;

	int xx, yy;

	for (xx = minx; xx <= maxx; xx++)
		if (ar[miny][xx] == 'X') break;
	tl = xx - minx;
	for (; xx <= maxx; xx++)
		if (ar[miny][xx] == '.') break;
	tr = maxx - xx + 1;

	for (xx = minx; xx <= maxx; xx++)
		if (ar[maxy][xx] == 'X') break;
	bl = xx - minx;
	for (; xx <= maxx; xx++)
		if (ar[maxy][xx] == '.') break;
	br = maxx - xx + 1;

	int dx = (maxx - minx) + 1;
	int dy = (maxy - miny) + 1;

	var = encode(dx, dy, tl, tr, bl, br);

	*mx = minx;
	*my = miny;
	return var;
}

int xflip(int state) {
	int sx = (state >> 20) & (0xF);
	int sy = (state >> 16) & (0xF);
	int tl = (state >> 12) & (0xF);
	int tr = (state >> 8) & (0xF);
	int tt = sx - tl - tr;
	int bl = (state >> 4) & (0xF);
	int br = (state) & (0xF);

	return encode(sx, sy, tr, tl, br, bl);
}


int flip(int state) {
	int sx = (state >> 20) & (0xF);
	int sy = (state >> 16) & (0xF);
	int tl = (state >> 12) & (0xF);
	int tr = (state >> 8) & (0xF);
	int tt = sx - tl - tr;
	int bl = (state >> 4) & (0xF);
	int br = (state) & (0xF);

	return encode(sy, sx, tl, bl, tr, br);
}

int rotate(int state) {
	int sx = (state >> 20) & (0xF);
	int sy = (state >> 16) & (0xF);
	int tl = (state >> 12) & (0xF);
	int tr = (state >> 8) & (0xF);
	int tt = sx - tl - tr;
	int bl = (state >> 4) & (0xF);
	int br = (state) & (0xF);

	return encode(sx, sy, br, bl, tr, tl);
}

int *divdiag_one(int state, int xx) {
	static int newstate[6];

	int sx = (state >> 20) & (0xF);
	int sy = (state >> 16) & (0xF);
	int tl = (state >> 12) & (0xF);
	int tr = (state >> 8) & (0xF);
	int tt = sx - tl - tr;
	int bl = (state >> 4) & (0xF);
	int br = (state) & (0xF);
	int bt = sx - bl - br;
	int lt = sy - tl - bl;
	int rt = sy - tr - br;

	int ntt[2], nlt[2], nlc[2];
	int ntc[2], nbr[2], nbl[2];

	if (xx < tl) {
		fflush(stdout);
		*(int *)(0xDEADBEEF) = 0;
	} else if (xx <= tl + tt) {
		ntt[0] = xx - tl;
		ntc[0] = 0;
		nbr[0] = 0;
	} else if (xx < sx + tr + 1) {
		ntt[0] = tt;
		ntc[0] = (xx - (tl + tt)) / 2;
		nbr[0] = (xx - (tl + tt)) % 2;
	} else {
		ntt[0] = tt;
		ntc[0] = tr;
		nbr[0] = xx - sx - tr;
	}

	if (xx <= lt + tl) {
		nlt[0] = xx - tl;
		nlc[0] = 0;
		nbl[0] = 0;
	} else if (xx < sy + bl + 1) {
		nlt[0] = lt;
		nlc[0] = (xx - (tl + lt)) / 2;
		nbl[0] = (xx - (tl + lt)) % 2;
	} else {
		nlt[0] = lt;
		nlc[0] = bl;
		nbl[0] = xx - sy - bl;
	}

	int cnt = 0;
	int xsub = 0, ysub = 0;
	for (int ii = 0; ii <= xx; ii++) {
		if (ii < tl + tt) cnt++;
		else if (ii >= tl + tt) {
			if ((ii - (tl + tt)) % 2) cnt++;
			else cnt--, ysub++;
		} else cnt--, ysub++;

		if (ii < tl + lt) cnt++;
		else if (ii >= tl + lt) {
			if ((ii - (tl + lt)) % 2) cnt++;
			else cnt--, xsub++;
		} else cnt--, xsub++;
	}
	newstate[1] = encode(cnt / 2, cnt / 2, (cnt / 2) - 1, 0, 0, (cnt / 2) - 1);
	newstate[1+3] = xsub << 4 | ysub;

	//printf("lastb: %x\n", tl+ntt[0]+ntc[0]-nlc[0]-nbl[0]-1);
	// collapse single corners
	int rmx = tl + ntt[0] + ntc[0];
	//int rmy = tl + nlt[0] + nlc[0];
	if (rmx > 0) {
		if (ntc[0] == rmx) ntc[0]--;
		if (nlc[0] == rmx) nlc[0]--;
		if (tl == rmx) tl--;
	}
	//printf("lasta: %x\n", tl+ntt[0]+ntc[0]-nlc[0]-nbl[0]-1);

	//if (tl + tl + ntt[0] + ntc[0] - nlc[0] - nbl[0] - 1 == tl + ntt[0] + ntc[0] + 1) {
	//	newstate[0] = 0;
	//} else {
		newstate[0] = encode(tl + ntt[0] + ntc[0], tl + nlt[0] + nlc[0], tl, ntc[0], nlc[0], max(0,tl + ntt[0] + ntc[0] - nlc[0] - nbl[0] - 1));
		newstate[0 + 3] = 0;
	//}

	return newstate;
}

int *divdiag(int state, int xx) {
	int sx = (state >> 20) & (0xF);
	int sy = (state >> 16) & (0xF);
	int rx = sx + sy - 2 - xx;

	static int newstate[6];

	int *temp_state = divdiag_one(state, xx);

	newstate[0] = temp_state[0];
	newstate[1] = temp_state[1];
	newstate[3] = temp_state[3];
	newstate[4] = temp_state[4];

	temp_state = divdiag_one(rotate(state), rx);

	newstate[2] = rotate(temp_state[0]);
	newstate[5] = (sx - ((newstate[2] >> 20) & (0xF))) << 4 | (sy - ((newstate[2] >> 16) & (0xF)));

	return newstate;
}
	

int *divvert(int state, int xx) {
	int sx = (state >> 20) & (0xF);
	int sy = (state >> 16) & (0xF);
	int tl = (state >> 12) & (0xF);
	int tr = (state >> 8) & (0xF);
	int tt = sx - tl - tr;
	int bl = (state >> 4) & (0xF);
	int br = (state) & (0xF);
	int bt = sx - bl - br;
	int lt = sy - tl - bl;
	int rt = sy - tr - br;

	static int newstate[6];

	int ta[2],tb[2],tc[2];
	int ba[2],bb[2],bc[2];

	if (xx >= tl + tt) {
		ta[0] = tl;
		tb[0] = tt;
		tc[0] = xx - (tl + tt);

		ta[1] = 0;
		tb[1] = 0;
		tc[1] = sx - xx - 1;
	} else if (xx >= tl && xx < tl + tt) {
		ta[0] = tl;
		tb[0] = xx - tl;
		tc[0] = 0;

		ta[1] = 0;
		tb[1] = tt - (xx - tl) - 1;
		tc[1] = tr;
	} else {
		ta[0] = xx;
		tb[0] = 0;
		tc[0] = 0;

		ta[1] = (tl - xx) - 1;
		tb[1] = tt;
		tc[1] = tr;
	}

	if (xx >= bl + bt) {
		ba[0] = bl;
		bb[0] = bt;
		bc[0] = xx - (bl + bt);

		ba[1] = 0;
		bb[1] = 0;
		bc[1] = sx - xx - 1;
	} else if (xx >= bl && xx < bl + bt) {
		ba[0] = bl;
		bb[0] = xx - bl;
		bc[0] = 0;

		ba[1] = 0;
		bb[1] = bt - (xx - bl) - 1;
		bc[1] = br;
	} else {
		ba[0] = xx;
		bb[0] = 0;
		bc[0] = 0;

		ba[1] = (bl - xx) - 1;
		bb[1] = bt;
		bc[1] = br;
	}

	// collapse single corners
	int rmx = xx;
	if (ta[0] == rmx) ta[0]--;
	if (tc[0] == rmx) tc[0]--;
	if (ba[0] == rmx) ba[0]--;
	if (bc[0] == rmx) bc[0]--;

	rmx = sx - xx - 1;
	if (ta[1] == rmx) ta[1]--;
	if (tc[1] == rmx) tc[1]--;
	if (ba[1] == rmx) ba[1]--;
	if (bc[1] == rmx) bc[1]--;

	newstate[0+3] = (0 << 4) | (tl - ta[0]);
	newstate[2+3] = ((xx + 1) << 4) | (tr - tc[1]);

	newstate[0] = encode(xx, lt + ta[0] + ba[0], ta[0], tc[0], ba[0], bc[0]);
	newstate[2] = encode(sx - xx - 1, rt + tc[1] + bc[1], ta[1], tc[1], ba[1], bc[1]);

	int hh = lt;
	if (tl > xx) hh += xx; else hh += tl;
	if (bl > xx) hh += xx; else hh += bl;
	if (xx >= sx - tr) hh -= (xx - (sx - tr) + 1);
	if (xx >= sx - br) hh -= (xx - (sx - br) + 1);
	newstate[1] = encode(1, hh, 0, 0, 0, 0);

	int nx = 0;
	if (tl > xx) nx = (tl - xx);
	if (xx >= sx - tr) nx = xx - sx + tr + 1;
		
	newstate[1+3] = (xx << 4) | nx;

	return newstate;
}

int solve(int state) {
	if (mm.find(state) != mm.end()) {
		return mm[state].first;
	}
	int sx = (state >> 20) & (0xF);
	int sy = (state >> 16) & (0xF);
	int tl = (state >> 12) & (0xF);
	int tr = (state >> 8) & (0xF);
	int bl = (state >> 4) & (0xF);
	int br = (state) & (0xF);

	if (sx == 0 || sy == 0) return 0;
	if (sx == 1 && sy == 1) return 0;

	printf("Solving %x\n", state);
	mm[state] = make_pair(REALLY_HUGE, REALLY_HUGE);

	int bx = -1;
	int ba = REALLY_HUGE;

	for (int xx = 0; xx < sx; xx++) {
		int newstate[6];
		memcpy(newstate, divvert(state, xx), sizeof(int)*6);
		//printf("xdiv at %d: %x(%d,%d) %x(%d,%d) %x(%d,%d)\n", xx,
		//	newstate[0], newstate[0+3] >>  4, (newstate[0+3] &0xF),
		//	newstate[1], newstate[1+3] >>  4, (newstate[1+3] &0xF),
		//	newstate[2], newstate[2+3] >>  4, (newstate[2+3] &0xF));
		int ts = max(max(solve(newstate[0]), solve(newstate[1])), solve(newstate[2])) + 1;
		if (ts == REALLY_HUGE) continue;
		if (ts < ba) {
			bx = 1 << 4;
			bx |= xx;
			ba = ts;
		}
	}

	for (int yy = 0; yy < sy; yy++) {
		int newstate[6];
		memcpy(newstate, divvert(flip(state), yy), sizeof(int)*6);
		//printf("ydiv at %d: %x(%d,%d) %x(%d,%d) %x(%d,%d)\n", yy,
		//	flip(newstate[0]), (newstate[0+3] &0xF), newstate[0+3] >>  4,
		//	flip(newstate[1]), (newstate[1+3] &0xF), newstate[1+3] >>  4,
		//	flip(newstate[2]), (newstate[2+3] &0xF), newstate[2+3] >>  4);
		int ts = max(max(solve(flip(newstate[0])), solve(flip(newstate[1]))), solve(flip(newstate[2]))) + 1;
		if (ts == REALLY_HUGE) continue;
		if (ts < ba) {
			bx = 3 << 4;
			bx |= yy;
			ba = ts;
		}
	}

	for (int xx = tl; xx < sx + sy - 1 - br; xx++) {
		int newstate[6];
		memcpy(newstate, divdiag(state, xx), sizeof(int) * 6);
		//printf("/div at %d: %x(%d,%d) %x(%d,%d) %x(%d,%d)\n", xx,
		//	newstate[0], newstate[0+3] >>  4, (newstate[0+3] &0xF),
		//	newstate[1], newstate[1+3] >>  4, (newstate[1+3] &0xF),
		//	newstate[2], newstate[2+3] >>  4, (newstate[2+3] &0xF));
		int ts = max(max(solve(newstate[0]), solve(newstate[1])), solve(newstate[2])) + 1;
		if (ts == REALLY_HUGE) continue;
		if (ts < ba) {
			bx = 2 << 4;
			bx |= xx;
			ba = ts;
		}
	}

	for (int xx = tr; xx < sx + sy - 1 - bl; xx++) {
		int newstate[6];
		memcpy(newstate, divdiag(xflip(state), xx), sizeof(int) * 6);
		int sd = sx - ((newstate[0] >> 20) & 0xF);
		//printf("\\div at %d: %x(%d,%d) %x(%d,%d) %x(%d,%d)\n", xx,
		//	xflip(newstate[0]), sd - (newstate[0+3] >> 4), (newstate[0+3] &0xF),
		//	xflip(newstate[1]), sd - (newstate[1+3] >> 4), (newstate[1+3] &0xF),
		//	xflip(newstate[2]), sd - (newstate[2+3] >> 4), (newstate[2+3] &0xF));
		int ts = max(max(solve(xflip(newstate[0])), solve(xflip(newstate[1]))), solve(xflip(newstate[2]))) + 1;
		if (ts == REALLY_HUGE) continue;
		if (ts < ba) {
			bx = 4 << 4;
			bx |= xx;
			ba = ts;
		}
	}


	printf("Solution at %x is div at %x (cost %d)\n", state, bx, ba);
	mm[state] = make_pair(ba, bx);

	return ba;
}

int real_solve(int state, int mx, int my) {

	while (solve(state) > 0) {
		int next_move = mm[state].second;
		int bt = (next_move >> 4);
		int bb = next_move & 0xF;
		int *newstate;
		int answer;

		printf("Real solve. I'm at %x (offset: %d, %d). Queries required: %d\n", state, mx, my, solve(state));

		switch(bt) {
			case 1:
				newstate = divvert(state, bb);
				answer = Oracle(mx + bb, my, 1);
				answer++;
				state = newstate[answer];
				mx += newstate[answer + 3] >> 4;
				my += newstate[answer + 3] & 0xF;
				break;
			case 3:
				newstate = divvert(flip(state), bb);
				answer = -Oracle(mx, my + bb, 3);
				answer++;
				state = flip(newstate[answer]);
				my += newstate[answer + 3] >> 4;
				mx += newstate[answer + 3] & 0xF;
				break;
			case 2:
				newstate = divdiag(state, bb);
				{
					int sx = (state >> 20) & 0xF;
					int dx, dy;
					if (bb >= sx) {
						dy = bb - sx + 1;
						dx = sx - 1;
					} else {
						dy = 0;
						dx = bb;
					}
		//printf("/div at %d: %x(%d,%d) %x(%d,%d) %x(%d,%d)\n", bb,
		//	newstate[0], newstate[0+3] >>  4, (newstate[0+3] &0xF),
		//	newstate[1], newstate[1+3] >>  4, (newstate[1+3] &0xF),
		//	newstate[2], newstate[2+3] >>  4, (newstate[2+3] &0xF));
					answer = Oracle(mx + dx, my + dy, 4);
					answer++;
					state = newstate[answer];
					mx += newstate[answer + 3] >> 4;
					my += newstate[answer + 3] & 0xF;
				}
				break;
			case 4:
				newstate = divdiag(xflip(state), bb);
				{
					int sx = (state >> 20) & 0xF;
					int dx, dy;
					if (bb >= sx) {
						dy = bb - sx + 1;
						dx = sx - 1;
					} else {
						dy = 0;
						dx = bb;
					}
					//printf("\\div at %d: %x(%d,%d) %x(%d,%d) %x(%d,%d)\n", bb,
					//	newstate[0], newstate[0+3] >>  4, (newstate[0+3] &0xF),
					//	newstate[1], newstate[1+3] >>  4, (newstate[1+3] &0xF),
					//	newstate[2], newstate[2+3] >>  4, (newstate[2+3] &0xF));
					answer = -Oracle(mx + sx - dx - 1, my + dy, 2);
					answer++;
					state = xflip(newstate[answer]);
					int sd = sx - ((newstate[answer] >> 20) & 0xF);
					mx += (sd - (newstate[answer + 3] >> 4));
					my += newstate[answer + 3] & 0xF;
				}
				break;
		}
	}
	Decide(mx, my);
	return 0;
}


int main() {
	freopen("news.in","r",stdin);
	scanf("%d%d ",&x,&y);
	for (int yy = y - 1; yy >= 0; yy--) {
		gets(ar[yy]);
	}
	for (int yy = 0; yy < y; yy++) {
		printf(ar[yy]);
		printf("\n");
	}
	int mx, my;

	int sq = canonar(&mx, &my);
	solve(sq);
	real_solve(sq, mx, my);

	printf("%x\n", sq);
}

